#ifndef _ENGINE_H_
#define _ENGINE_H_
#include <setjmp.h>
#include "wire.h"
#include "box.h"

/* these two constants MUST agree with the compier, otherwise things will die! */
#ifdef UNBOXED
#define BUILTIN_EXCEPTIONS_OFFSET 1
#else 
#define BUILTIN_EXCEPTIONS_OFFSET 8
#define USER_CONSTRUCTORS_OFFSET 32
#endif /*UNBOXED*/

typedef enum {
 StackOverflow = BUILTIN_EXCEPTIONS_OFFSET, 
 HeapOverflow, 
 WireOverflow,
 MatchFail, 
 DivisionByZero, 
 InternalError,
 IndexOutOfBounds, 
 
 BoxWithin,
 BoxTimeout,
 WireWithin,
 WireTimeout,
 ExpressionWithin,
 EndOfFile,

 SoftStackOverflow,
 SoftHeapOverflow
} exception_e;

/*
 * following are generated by the compiler
 */
void engineInit();
void engineRun();

extern uint16 wireCnt();
extern uint16 boxCnt();
extern wire_t wire[];
extern box_t box[];
extern STACK stack[];
extern HEAP heap[];


/* to support 'debugging' should implement these to return something other than NULL */
const char *getBoxName(box_t *b);
const char *getWireName(wire_t *w);
const char *getConName(uint16 c);




extern uint32 nSchedules; /* 0 = unlimited */
extern uint32 schedules;



// 64-bit issues. NRS 02-Dec-08
//extern uint32 pc;
#if SIZEOF_VOID_P==4
extern uint32 pc;
#else
#if SIZEOF_VOID_P==8
extern uint64 pc;
#else
#error Unknown SIZEOF_VOID_P
#endif
#endif

extern SP sp, mp, rp, fp, slp;
extern SP spLimit;
extern SP spSoftLimit;

extern HEAP *hp;
extern HEAP *hpLimit;
extern HEAP *hpSoftLimit;
extern box_t *currentBox;

extern jmp_buf excJmp;

/*
 * no-op instruction, I don't know why... maybe it's so you can attach a debugger to it or something. blaaarghhhhh.
 *
 */
void noOp();


/* 
 * User defined exceptions
 * - typically follow this with a "goto __humeHandler;" 
 */
void setException(exception_e exc, HEAP value);

/* 
 * RTS and Async exceptions which have Tup0 as the argument.
 * - WARNING, WILL NOT RETURN (uses longjmp/setjmp)
 */
void throwException(exception_e exc);


/*
 * Copy one structure into another
 * - must be same type and shape
 */
void unsafeAssignment(HEAP *dest, HEAP *source);

/*
 * closure function call
 */
boolean callVar(SP lsp, uint32 extras, int ret);

/* pascal style string equivalence */
boolean pstreq(char *str1, uint16 len1, char *str2, uint16 len2);

#ifndef UNBOXED
/*
 * Compare two structures based on given flags
 * - Note: flags eq && lt should not both be FALSE.
 * - PS: this may not be the fastest code, but it is the most compact.
 *
 */
boolean cmpPrim(HEAP *arg1, HEAP *arg2, boolean eq, boolean lt);
#else
typedef enum {RES_LESS, RES_NEQ, RES_EQ} res_e;
#endif /*UNBOXED*/

#ifdef NDEBUG
/* No stack cheaks anymore... */
#define inc_sp() (sp++)
#define inc_spn(x) (sp+=x)
#else
SP inc_sp();
SP inc_spn(uint32 n);
#endif /*NDEBUG*/


/*
 * Stack frame definitions
 */
#define RET_PC() (fp-4)
#define RET_VALUE() (fp-4)
#define PREV_FN() (fp-3)
#define PREV_LINK() (fp-2)
#define PREV_FP() (fp-1)
#define PREV_LINKF(fp) (fp-2)
#define PREV_FPF(fp) (fp-1)
#define ARG(n) ((fp-5)-n)
#define VAR(n) VARF(fp,n)
#define VARF(fp,n) (fp+n)



/*
 * Computed goto support
 */

#ifdef __GNUC__
/* fast and constant time per branch*/
#define PC_ADDR(n, x) ((uint32)&& n##_##x)
#define PC_GOTO(n) goto *((void*)pc)

#else
/* slow and varying time per branch*/
#define PC_TABLE
#define PC_ADDR(n, x) x
#define PC_GOTO(n) goto n;

#endif 





/*
 * current position of the fair rule matching
 */
extern uint16 fairPos; 

/*
 * initilize the rule to 0..size-1 
 */
 
void ruleInit(uint16 *rule, uint16 size);
/* 
 * push the current rule to the end of the LRU queue 
 */
 
void rulePush(uint16 *rule, uint16 end);



/* 
 * Returns the ord() of discriminated unions
 *  - the real value (if ever exposed to Hume) is (ord()<<2)-1
 *  - for simple wrapped unions you need to pass the tag of the wrapped varient, otherwise pass 0
 */

uint32 ordFor(HEAP v, uint32 tag);

#endif /*_ENGINE_H_*/
